// $Id: Application.cpp,v 1.33 2010/07/22 22:03:39 hpereira Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* software; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA  02111-1307 USA
*
*
*******************************************************************************/

/*!
   \file    Application.cc
   \brief   application Main Window singleton object
   \author  Hugo Pereira
   \version $Revision: 1.33 $
   \date    $Date: 2010/07/22 22:03:39 $
*/

#include <QApplication>
#include <QMessageBox>

#include "ConfigurationDialog.h"
#include "Config.h"
#include "Debug.h"
#include "File.h"
#include "IconEngine.h"
#include "Icons.h"
#include "Application.h"
#include "XmlOptions.h"
#include "InformationDialog.h"
#include "QuestionDialog.h"
#include "TaskBarBase.h"
#include "TaskBarConfiguration.h"
#include "TaskBarDialog.h"
#include "TaskBarFixed.h"
#include "TaskBarMinimum.h"
#include "TaskBarParabolic.h"
#include "TaskParser.h"

#include "CompositeEngine.h"

using namespace std;
using namespace Qt;

//____________________________________________
void Application::usage( void )
{

  Debug::Throw(0) << "Usage : taskbar [options] [configuration file]" << endl;
  SERVER::ApplicationManager::commandLineParser().usage();
  return;

}

//____________________________________________
Application::Application( CommandLineArguments arguments ) :
  BaseApplication( 0, arguments ),
  Counter( "Application" ),
  MAIN_TITLE( "Taskbar" )
{
  Debug::Throw( "Application::Application.\n" );
  connect( qApp, SIGNAL( aboutToQuit() ), this, SLOT( _aboutToQuit() ) );
  qApp->setQuitOnLastWindowClosed( false );
}

//____________________________________________
Application::~Application( void )
{
  Debug::Throw( "Application::~Application.\n" );
}

//____________________________________________
bool Application::realizeWidget( void )
{
  Debug::Throw( "Application::realizeWidget.\n" );

  // check if the method has already been called.
  if( !BaseApplication::realizeWidget() ) return false;

  // rename about action
  aboutAction().setText( "About &taskbar" );
  configurationAction().setText( "Configure &Transparent Taskbar ..." );

  // change close action shortcut
  closeAction().setShortcut( QKeySequence() );

  add_taskbar_action_ = new QAction( IconEngine::get( ICONS::ADD ), "&Add Taskbar", this );
  connect( add_taskbar_action_, SIGNAL( triggered() ), SLOT( _addTaskBar() ) );

  remove_taskbar_action_ = new QAction( IconEngine::get( ICONS::REMOVE ), "&Remove Taskbar", this );
  connect( remove_taskbar_action_, SIGNAL( triggered() ), SLOT( _removeTaskBar() ) );

  // load taskbars
  _readTaskBars();

  // configuration
  emit configurationChanged();

  // show all taskbars
  for( TaskBarBase::List::const_iterator iter = taskbars_.begin(); iter != taskbars_.end(); iter++ )
  { (*iter)->show(); }

  return true;

}

//___________________________________________________
void Application::_configuration( void )
{
  Debug::Throw( "Application::_configuration.\n" );

  emit saveConfiguration();

  ConfigurationDialog dialog;
  connect( &dialog, SIGNAL( configurationChanged() ), this, SIGNAL( configurationChanged() ) );
  dialog.centerOnWidget( qApp->activeWindow() );
  dialog.exec();

}

//_______________________________________________
void Application::_addTaskBar( void )
{

  Debug::Throw( "Application::_addTaskBar.\n" );
  TaskBarDialog dialog( currentTaskBar() );
  if( !dialog.exec() ) return;

  // create new taskbar, put in list and write to file
  TaskBarBase *bar = _newTaskBar( dialog.taskBarConfiguration().attributes() );
  taskbars_.push_back( bar );
  bar->updateConfiguration();
  bar->show();
  _writeTaskBars();

  // update actions
  _updateActions();

}

//_______________________________________________
void Application::_editTaskBar( const TaskBarAttributes& attributes )
{

  Debug::Throw( "Application::_editTaskBar.\n" );

  // retrieve current Taskbar
  TaskBarBase* bar( currentTaskBar() );
  Debug::Throw( "Application::_editTaskBar - got bar.\n" );

  // check if attributes are changed
  if( attributes == bar->attributes() )
  {
    Debug::Throw( "Application::_editTaskBar - attributes are unchanged.\n" );
    return;
  }

  // update current task bar (the TaskAttributesList is unchanged)
  TaskAttributes::List task_list( bar->attributes().taskAttributesList() );
  _setCurrentAttributes( attributes );
  _currentAttributes().setTaskAttributesList( task_list );

  Debug::Throw( "Application::_editTaskBar - attributes set.\n" );

  // if taskbar mode differ, delete taskbar, create a new one and replace
  // otherwise, just update attributes
  if( _currentAttributes().mode() != bar->attributes().mode() )
  {

    _removeTaskBar( false );
    bar = _newTaskBar( _currentAttributes() );
    bar->updateConfiguration();
    taskbars_.push_back( bar );
    bar->show();

  } else { bar->setAttributes( _currentAttributes() ); }

  // save taskbars
  _writeTaskBars();

  Debug::Throw( "Application::_editTaskBar - done.\n" );
  return;

}

//_______________________________________________
void Application::_removeTaskBar( const bool& ask )
{
  Debug::Throw( "Application::_removeTaskBar.\n" );

  // retrieve current Taskbar
  TaskBarBase* bar( currentTaskBar() );

  if( ask )
  {
    QString buffer;
    QTextStream( &buffer ) << "Remove taskbar " <<  _currentAttributes().name() << " ?";
    if( !QuestionDialog( bar, buffer ).exec() ) return;
  }

  // delete taskbars matching attributes
  TaskBarBase::List tmp;
  for( TaskBarBase::List::iterator iter = taskbars_.begin(); iter != taskbars_.end(); iter++ )
  if( (*iter)->attributes().id() == _currentAttributes().id() ) (*iter)->deleteLater();
  else tmp.push_back( *iter );
  taskbars_ = tmp;

  // write new set of taskbars to file
  _writeTaskBars();

  // update actions
  _updateActions();

}

//_______________________________________________
void Application::_readTaskBars( void )
{
  QString filename( XmlOptions::get().raw("XML_FILE" ) );

  Debug::Throw() << "Application::_readTaskBars - file: " << filename << endl;

  // read tasks from db file
  TaskParser parser;
  parser.read( filename );

  // check errors
  if( parser.xmlError() )
  {
    QString buffer;
    QTextStream what( &buffer );
    what << "An error occured while parsing file." << endl;
    what << parser.xmlError();
    InformationDialog( 0, buffer ).exec();
  }

  // local copy
  const TaskBarAttributes::List& attributes( parser.attributes() );
  TaskBarAttributes::List::const_iterator iter = attributes.begin();
  TaskBarBase::List::iterator bar_iter = taskbars_.begin();

  // update existing taskbars
  while( iter != attributes.end() && bar_iter != taskbars_.end() )
  {

    // check if zoom mode match
    if( (*bar_iter)->attributes().mode() != iter->mode() )
    {

      (*bar_iter)->deleteLater();
      *bar_iter = _newTaskBar( *iter );

    } else {

      (*bar_iter)->setAttributes( *iter );

    }

    (*bar_iter)->show();
    iter++;
    bar_iter++;

  }

  if( iter != attributes.end() )
  {

    // additional taskbar if needed
    for( ; iter != attributes.end(); iter++ )
    {
      TaskBarBase *bar = _newTaskBar( *iter );
      taskbars_.push_back( bar );
    }

  } else if( bar_iter != taskbars_.end() ) {

    // remove excedent taskbar if needed
    for( TaskBarBase::List::iterator bar_iter2 = bar_iter; bar_iter2 != taskbars_.end(); bar_iter2++ )
    { (*bar_iter2)->deleteLater(); }

    taskbars_.erase( bar_iter, taskbars_.end() );

  }

  // update actions
  _updateActions();

  return;

}

//_______________________________________________
void Application::_writeTaskBars( bool check ) const
{

  QString filename( XmlOptions::get().raw("XML_FILE" ) );
  Debug::Throw() << "Application::_writeTaskBars - filename: " << filename << endl;

  // if taskbars have not been initialized, do nothing
  if( !_realized() )
  {
    Debug::Throw( "Application::_writeTaskBars - application not realized. Aborted.\n" );
    return;
  }

  // check taskbar
  if( taskbars_.empty() )
  {
    Debug::Throw( "Application::_writeTaskBars - taskbar list empty.\n" );
    return;
  }

  // try open XML file
  File file( filename );

  TaskParser parser;
  for( TaskBarBase::List::const_iterator iter = taskbars_.begin(); iter != taskbars_.end(); iter++ )
  {
    (*iter)->synchronizeAttributes();
    parser.addAttributes( (*iter)->attributes() );
  }

  bool written( parser.write( file ) );
  if( (!written) && check )
  {
    QString buffer;
    QTextStream( &buffer ) << "unable to write to file " << file;
    InformationDialog( 0, buffer ).exec();
  }

  return;

}

//_______________________________________________
void Application::_aboutToQuit( void )
{
  Debug::Throw( "Application::_AboutToQuit.\n" );
  _writeTaskBars();
}

//________________________________________________
bool Application::_processCommand( SERVER::ServerCommand command )
{

  Debug::Throw( "Application::_processCommand.\n" );
  if( BaseApplication::_processCommand( command ) ) return true;
  if( command.command() == SERVER::ServerCommand::RAISE )
  { return true; }

  return false;

}

//_______________________________________________
TaskBarBase* Application::_newTaskBar( const TaskBarAttributes& attributes )
{

  Debug::Throw( "Application::_newTaskBar.\n" );

  TaskBarBase* bar = 0;
  switch ( attributes.mode() )
  {

    case TaskBarAttributes::FIXED:
    bar = new TaskBarFixed();
    break;


    case TaskBarAttributes::MINIMUM:
    bar = new TaskBarMinimum();
    break;

    case TaskBarAttributes::OSX:
    bar = new TaskBarParabolic();
    break;

    default: assert(0);
    break;

  }

  connect( bar, SIGNAL( modified() ), this, SLOT( _writeTaskBars() ) );
  connect( bar, SIGNAL( taskBarEntered( const TaskBarAttributes& ) ), SLOT( _setCurrentAttributes( const TaskBarAttributes& ) ) );
  connect( bar, SIGNAL( attributesChanged( const TaskBarAttributes& ) ), SLOT( _editTaskBar( const TaskBarAttributes& ) ) );
  connect( bar, SIGNAL( taskListChanged() ), SLOT( _writeTaskBars() ) );

  bar->setWindowIcon( QPixmap( File( XmlOptions::get().raw( "ICON_PIXMAP" ) ).expand() ) );
  bar->setWindowTitle( MAIN_TITLE );
  bar->setAttributes( attributes, false );

  // close action is added to the bar to enable the shortcut
  bar->addAction( &closeAction() );

  return bar;

}

//_______________________________________________
TaskBarBase* Application::_taskBar( const TaskBarAttributes& attributes ) const
{
  Debug::Throw( "Application::_taskBar.\n" );
  TaskBarBase::List::const_iterator iter( find_if( taskbars_.begin(), taskbars_.end(), TaskBarBase::SameAttributesFTor( attributes ) ) );
  assert( iter != taskbars_.end() );

  Debug::Throw( "Application::_taskBar - found.\n" );
  return *iter;

}


//_______________________________________________
void Application::_updateActions( void ) const
{
  Debug::Throw( "Application::_updateAction.\n" );
  removeTaskBarAction().setEnabled( taskbars_.size() > 1 );
}
