// $Id: TaskBarParabolic.cpp,v 1.1 2010/07/22 22:03:39 hpereira Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* software; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place, Suite 330, Boston, MA  02111-1307 USA
*
*
*******************************************************************************/

/*!
\file    TaskpainterOsx.cc
\brief   Osx style task painting
\author  Hugo Pereira
\version $Revision: 1.1 $
\date    $Date: 2010/07/22 22:03:39 $
*/

#include <algorithm>
#include <cmath>
#include <QApplication>
#include <QDesktopWidget>

#include "QtUtil.h"
#include "RoundedRegion.h"
#include "SvgEngine.h"
#include "TaskBarParabolic.h"
#include "Util.h"
#include "XmlOptions.h"

#include "CompositeEngine.h"

using namespace std;

//____________________________________________________
TaskBarParabolic::TaskBarParabolic( QWidget* parent ):
    TaskBarBase( parent ),
    maxIndex_(0),
    index_(0),
    direction_( TaskButton::NONE )
{ Debug::Throw( "TaskBarParabolic::TaskBarParabolic.\n" ); }

//____________________________________________________
void TaskBarParabolic::_updateRootPixmap( void )
{
    
    _setRootPixmapChanged( false );
    
    if( Debug::level() >= 1 )
    {
        // solid background
        _setRootPixmap( QPixmap( size() ) );
        _rootPixmap().fill( palette().color( backgroundRole() ) );
        
    } else if( TRANSPARENCY::CompositeEngine::get().isEnabled() ) {
        
        _setRootPixmap( QPixmap( size() ) );
        QPainter painter( &_rootPixmap() );
        painter.setRenderHints(QPainter::SmoothPixmapTransform);
        painter.setCompositionMode(QPainter::CompositionMode_Source );
        painter.fillRect( _rootPixmap().rect(), Qt::transparent);
        
    } else {
        
        _setRootPixmap( QPixmap::grabWindow(
            QApplication::desktop()->winId(),
            geometry().x(), geometry().y(),
            geometry().width(), geometry().height() ) );
        
    }
    
}

//___________________________________________________
void TaskBarParabolic::_updateBackgroundPixmap( void )
{
    Debug::Throw( "TaskBarParabolic::TaskBarFixed.\n" );
    
    if( _rootPixmapChanged() )
    {
        // there is no need to update the root pixmap when composite engine is off,
        // unless during resize events, for which the method is called directly
        if( !TRANSPARENCY::CompositeEngine::get().isEnabled() ) _updateRootPixmap();
        else _setRootPixmapChanged( false );
    }
    
    TaskBarBase::_updateBackgroundPixmap();
    
    // need to draw svg on background pixmap if button list is empty
    // due to the way empty taskbars are implemented
    if( _buttons().empty() && _useSvg() && SVG::SvgEngine::get().isValid() )
    {
        QPainter painter( &_backgroundPixmap() );
        painter.setRenderHints(QPainter::SmoothPixmapTransform);
        painter.setCompositionMode(QPainter::CompositionMode_Source );
        painter.drawPixmap( QPoint(0,0),  SVG::SvgEngine::get().get( size() ) );
    }
    
}

//______________________________________________
void TaskBarParabolic::resizeEvent( QResizeEvent* event )
{
    _updateRootPixmap();
    
    /*
    need to bypass TaskBarBase::resizeEvent, in order not to have
    the entire background blurred, when enabled 
    */
    if( attributes().flag( TaskBarAttributes::ROUND_CORNERS ) ) _roundCorners();
    TransparentWidget::resizeEvent( event );
}

//________________________________________
void TaskBarParabolic::enterEvent( QEvent *event )
{
    
    Debug::Throw( "TaskBarParabolic::enterEvent.\n" );
    if( _mouseButton() == Qt::NoButton ) _setDirection( TaskButton::BIGGER );
    TaskBarBase::enterEvent( event );
    
}

//___________________________________________________
void TaskBarParabolic::leaveEvent( QEvent *event )
{
    
    Debug::Throw( "TaskBarParabolic::leaveEvent.\n" );
    _setDirection( TaskButton::SMALLER );
    TaskBarBase::leaveEvent( event );
    
}

//___________________________________________________
void TaskBarParabolic::mouseMoveEvent( QMouseEvent *event )
{
    Debug::Throw( "TaskBarParabolic::mouseMoveEvent.\n" );
    TaskBarBase::mouseMoveEvent( event );
    update();
}

//_____________________________________________________________
void TaskBarParabolic::_updateMinimumSize( void )
{
    
    Debug::Throw( "TaskBarParabolic::_updateMinimumSize.\n" );
    
    // compute minimum size
    QSize minimum_size;
    
    // compute visible rect
    const TaskBarAttributes::Layout& layout ( attributes().layout() );
    if( layout & TaskBarAttributes::VERTICAL )
    {
        
        for( TaskButton::List::const_iterator iter = _buttons().begin(); iter != _buttons().end(); iter++ )
        {
            minimum_size.setHeight( minimum_size.height()+(*iter)->minimumSize().height()+_spacing() );
            minimum_size.setWidth( max<int>( minimum_size.width(), (*iter)->minimumSize().width() + _spacing() + 2*_margin() ) );
        }
        
        minimum_size.setHeight( minimum_size.height() + 2*_margin() );
        
    } else if( layout & TaskBarAttributes::HORIZONTAL ) {
        
        for( TaskButton::List::const_iterator iter = _buttons().begin(); iter != _buttons().end(); iter++ )
        {
            minimum_size.setWidth( minimum_size.width() + (*iter)->minimumSize().width()+_spacing() );
            minimum_size.setHeight( max<int>( minimum_size.height(), (*iter)->minimumSize().height() + _spacing() + 2*_margin() ) );
        }
        
        minimum_size.setWidth( minimum_size.width() + 2*_margin() );
        
    } else {
        
        unsigned int column_index(0);
        unsigned int line_index(0);
        
        // compute column width/line height
        vector<unsigned int> column_width( _numColumns(), 0 );
        vector<unsigned int> line_height( _numLines(), 0 );
        for( TaskButton::List::iterator iter = _buttons().begin(); iter != _buttons().end(); iter++ )
        {
            
            line_height[line_index] = max( line_height[line_index], (*iter)->minimumSize().height() + _spacing() );
            column_width[column_index] = max( column_width[column_index], (*iter)->minimumSize().width() + _spacing() );
            column_index++;
            if( column_index >= _numColumns() ) {
                column_index = 0;
                line_index++;
            }
            
        }
        
        int line_width(0);
        column_index = 0;
        line_index = 0;
        for( TaskButton::List::iterator iter = _buttons().begin(); iter != _buttons().end(); iter++ )
        {
            
            line_width += column_width[column_index];
            column_index++;
            
            // change column
            if( column_index >= _numColumns() )
            {
                minimum_size.setWidth( max( minimum_size.width(), line_width ) );
                minimum_size.setHeight( minimum_size.height()+line_height[line_index] );
                
                line_width = 0;
                column_index = 0;
                line_index++;
                
            }
            
        }
        
        minimum_size.setHeight( minimum_size.height() + line_height.back() + 2*_margin() );
        minimum_size.setWidth( minimum_size.width() + 2*_margin() );
        
    }
    
    // update
    _setMinimumSize( minimum_size );
    return;
    
}

//_____________________________________________________________
void TaskBarParabolic::_updateMinimumRectPosition( void )
{
    
    Debug::Throw( "TaskBarParabolic::_updateMinimumRectPosition.\n" );
    
    QRect minimum_rect( _minimumRect() );
    QPoint offset;
    const TaskBarAttributes::Layout& layout ( attributes().layout() );
    if( layout & TaskBarAttributes::VERTICAL ) {
        
        offset.setY( ( _maximumSize().height() - minimum_rect.height() )/2 );
        if( layout & TaskBarAttributes::BOTTOM_RIGHT ) offset.setX( _maximumSize().width() - minimum_rect.width() );
        else if( layout & TaskBarAttributes::CENTER ) offset.setX( ( _maximumSize().width() - minimum_rect.width() )/2 );
        
    } else if( layout & TaskBarAttributes::HORIZONTAL ) {
        
        offset.setX( ( _maximumSize().width() - minimum_rect.width() )/2 );
        if( layout & TaskBarAttributes::BOTTOM_RIGHT ) offset.setY( _maximumSize().height() - minimum_rect.height() );
        else if( layout & TaskBarAttributes::CENTER ) offset.setY( ( _maximumSize().height() - minimum_rect.height() )/2 );
        
    } else  {
        
        offset.setX( ( _maximumSize().width() - minimum_rect.width() )/2 );
        offset.setY( ( _maximumSize().height() - minimum_rect.height() )/2 );
        
    }
    
    minimum_rect.moveTo( offset );
    _setMinimumRect( minimum_rect );
    
}

//_____________________________________________________________
void TaskBarParabolic::_updateVisibleRect( void )
{
    
    Debug::Throw( "TaskBarParabolic::_updateVisibleRect.\n" );
    
    QRect visible_rect;
    const TaskBarAttributes::Layout& layout ( attributes().layout() );
    if( layout & TaskBarAttributes::VERTICAL )
    {
        
        // width is from minimum size
        visible_rect.setWidth( _minimumSize().width() );
        
        // when compositing is enabled one can shrink the visible rect
        // it does not work otherwise, due to 'echoes' in the root pixmap
        if( TRANSPARENCY::CompositeEngine::get().isEnabled() && _buttons().size() > 1 )
        {
            
            // get bounding height and y offset
            bool first( true );
            for( TaskButton::List::iterator iter = _buttons().begin(); iter != _buttons().end(); iter++ )
            {
                visible_rect.setTop( first ? (*iter)->rect().top() : min( visible_rect.top(), (*iter)->rect().top() ) );
                visible_rect.setBottom( max( visible_rect.bottom(), (*iter)->rect().bottom() ) );
                first = false;
            }
            
            visible_rect.setTop( visible_rect.top() - _margin() );
            visible_rect.setBottom( visible_rect.bottom() + _margin() );
            
        } else {
            
            visible_rect.setHeight( _maximumSize( _index() ).height() );
            visible_rect.moveTop( (height() - _maximumSize( _index() ).height())/2 );
            
        }
        
        if( layout & TaskBarAttributes::BOTTOM_RIGHT ) visible_rect.moveLeft(  _maximumSize().width() - visible_rect.width() );
        else if( layout & TaskBarAttributes::CENTER ) visible_rect.moveLeft( ( _maximumSize().width() - visible_rect.width() )/2 );
        
    } else if( layout & TaskBarAttributes::HORIZONTAL ) {
        
        // height is from minimum size
        visible_rect.setHeight( _minimumSize().height() );
        
        // when compositing is enabled one can shrink the visible rect
        // it does not work otherwise, due to 'echoes' in the root pixmap
        if( TRANSPARENCY::CompositeEngine::get().isEnabled() && _buttons().size() > 1  )
        {
            
            // get bounding height and y offset
            bool first( true );
            for( TaskButton::List::iterator iter = _buttons().begin(); iter != _buttons().end(); iter++ )
            {
                visible_rect.setLeft( first ? (*iter)->rect().left() : min( visible_rect.left(), (*iter)->rect().left() ) );
                visible_rect.setRight( max( visible_rect.right(), (*iter)->rect().right() ) );
                first = false;
            }
            
            visible_rect.setLeft( visible_rect.left() - _margin() );
            visible_rect.setRight( visible_rect.right() + _margin() );
            
        } else {
            
            visible_rect.setWidth( _maximumSize( _index() ).width() );
            visible_rect.moveLeft( (width() - _maximumSize( _index() ).width())/2 );
            
        }
        
        if( layout & TaskBarAttributes::BOTTOM_RIGHT ) visible_rect.moveTop(  _maximumSize().height() - visible_rect.height() );
        else if( layout & TaskBarAttributes::CENTER ) visible_rect.moveTop( ( _maximumSize().height() - visible_rect.height() )/2 );
        
    } else {
        
        visible_rect.setSize( _minimumSize() );
        visible_rect.moveTo( ( _maximumSize().width() - visible_rect.width() )/2, ( _maximumSize().height() - visible_rect.height() )/2 );
        
    }
    
    // add offset (for auto-hide)
    visible_rect.translate( _offset() );
    
    // in case of autoHide, one make sure there is at least
    // a thin band drawn at the bottom, that match the hidden geometry
    if( attributes().flag( TaskBarAttributes::AUTO_HIDE) )
    {
        
        const TaskBarAttributes::Layout& layout ( attributes().layout() );
        if( layout & TaskBarAttributes::HORIZONTAL )
        {
            if( layout & TaskBarAttributes::TOP_LEFT )  visible_rect.setBottom( max(  visible_rect.bottom(), 3 ) );
            else visible_rect.setTop( min( visible_rect.top(), height()-3 ) );
        } else {
            if( layout & TaskBarAttributes::TOP_LEFT ) visible_rect.setRight( max( visible_rect.right(), 3 ) );
            else visible_rect.setLeft( min( visible_rect.left(), width()-3 ) );
        }
        
    }
    
    // update blur areas if visibleRect is changed
    if( _setVisibleRect( visible_rect ) && _blurEnabled() )
    {
        
        static const int margin = 10;
        _updateBlurRegion( _visibleRect().translated( mapTo( window(), QPoint(0,0) ) ).adjusted( margin, margin, -margin, -margin ) );
        
    }
    
    return;
    
}


//_____________________________________________________________
bool TaskBarParabolic::_updateGeometry( Action action )
{
    
    Debug::Throw() << "TaskBarParabolic::_updateGeometry - action: " << action << endl;
    
    if( action == NONE ) return false;
    
    // minimum action
    if( action == MINIMUM )
    {
        
        setMask( _minimumRect() );
        if( !TRANSPARENCY::CompositeEngine::get().isEnabled() ) _setRootPixmapChanged( true );
        _setAction( NONE );
        return false;
        
    }
    
    // enter action
    if( action == ENTER ) {
        
        if( _rootPixmapChanged() )
        {
            // there is no need to update the root pixmap when composite engine is off,
            // unless during resize events, for which the method is called directly
            if( !TRANSPARENCY::CompositeEngine::get().isEnabled() ) _updateRootPixmap();
            else _setRootPixmapChanged( false );
        }
        
        
        if( !TaskBarBase::_updateGeometry( action ) )
        {
            
            _setAction( NONE );
            setMask( rect() );
            return false;
            
        } else return true;
        
    }
    
    // leave action
    if( action == LEAVE )
    {
        
        if( !_rootPixmapChanged() )
        {
            setMask( _minimumRect() );
            _setRootPixmapChanged( true );
        }
        
        if( !TaskBarBase::_updateGeometry( action ) )
        {
            
            _setAction( NONE );
            return false;
            
        } else return true;
        
    }
    
    return false;
    
}

//_________________________________________________________________
void TaskBarParabolic::_updateMaximumSize( void )
{
    
    Debug::Throw( "TaskBarParabolic::_updateMaximumSize.\n" );
    
    // maxumum size and index
    _setMaximumIndex( (*min_element( _buttons().begin(), _buttons().end(), TaskButton::MaxIndexFTor() ))->maximumIndex() );
    _maximumButtonSize().setWidth( (*max_element( _buttons().begin(), _buttons().end(), TaskButton::MaxWidthFTor() ))->maximumSize().width() + _spacing() );
    _maximumButtonSize().setHeight( (*max_element( _buttons().begin(), _buttons().end(), TaskButton::MaxHeightFTor() ))->maximumSize().height() + _spacing() );
    
    // reset button indices
    _setIndex(0);
    _placeButtons();
    
    // reset mouse offset
    _setMouseOffset( QPoint() );
    _clearMaximumSizes();
    _clearMidPositions();
    for( unsigned int index = 0; index < _maximumIndex(); index++ )
    {
        
        QSize maximum_size;
        
        // retrieve taskbar layout
        const TaskBarAttributes::Layout& layout( attributes().layout() );
        if( layout & TaskBarAttributes::HORIZONTAL ) maximum_size.setHeight( _maximumButtonSize().height() );
        else if( layout & TaskBarAttributes::VERTICAL ) maximum_size.setWidth( _maximumButtonSize().width() );
        
        // store the mid position
        QPoint mid_position;
        int counts( 0 );
        
        // set index to max value
        _setIndex( index );
        
        // increment mouse position over the entire taskbar to derive the maximum size
        for( TaskButton::List::iterator iter = _buttons().begin(); iter != _buttons().end(); iter++ )
        {
            
            // set mouse position
            // the use of 'center' position when there is only one button is to fix display
            // why it is not usable for button size larger than 1 is unclear.
            // whether the second works for other orientations is also unclear
            // -> to be fixed
            if( _buttons().size() == 1 ) _setMousePosition( (*iter)->rect().center() );
            else _setMousePosition( (*iter)->rect().topLeft() );
            
            // place buttons
            _placeButtons();
            
            if( layout & TaskBarAttributes::HORIZONTAL )
            {
                
                // get bounding rect
                int bounding_size(0);
                for( TaskButton::List::iterator sub_iter = _buttons().begin(); sub_iter != _buttons().end(); sub_iter++ )
                { bounding_size = max( bounding_size, (*sub_iter)->rect().right() ); }
                
                // update maximum size
                if( bounding_size > maximum_size.width() )
                {
                    maximum_size.setWidth( bounding_size );
                    mid_position = _mousePosition();
                    counts = 1;
                    
                } else if( bounding_size == maximum_size.width() ) {
                    
                    mid_position += _mousePosition();
                    counts++;
                    
                }
                
            } else if( layout & TaskBarAttributes::VERTICAL ) {
                
                // get bounding rect
                int bounding_size(0);
                for( TaskButton::List::iterator iter = _buttons().begin(); iter != _buttons().end(); iter++ )
                { bounding_size = max( bounding_size, (*iter)->rect().bottom() ); }
                
                // update maximum size
                if( bounding_size > maximum_size.height() )
                {
                    maximum_size.setHeight( bounding_size );
                    mid_position = _mousePosition();
                    counts = 1;
                    
                } else if( bounding_size == maximum_size.height() ) {
                    
                    mid_position += _mousePosition();
                    counts++;
                    
                }
                
            } else {
                
                // get bounding rect
                for( TaskButton::List::iterator sub_iter = _buttons().begin(); sub_iter != _buttons().end(); sub_iter++ )
                {
                    maximum_size.setWidth( max( maximum_size.width(), (*sub_iter)->rect().left() ) );
                    maximum_size.setHeight( max( maximum_size.height(), (*sub_iter)->rect().bottom() ) );
                }
                
                // set mouse position
                _setMousePosition( (*iter)->rect().center() );
                _placeButtons();
                
                // get bounding rect
                for( TaskButton::List::iterator sub_iter = _buttons().begin(); sub_iter != _buttons().end(); sub_iter++ )
                {
                    maximum_size.setWidth( max( maximum_size.width(), (*sub_iter)->rect().left() ) );
                    maximum_size.setHeight( max( maximum_size.height(), (*sub_iter)->rect().bottom() ) );
                }
                
                // set mouse position
                _setMousePosition( (*iter)->rect().bottomRight() );
                _placeButtons();
                
                // get bounding rect
                for( TaskButton::List::iterator sub_iter = _buttons().begin(); sub_iter != _buttons().end(); sub_iter++ )
                {
                    maximum_size.setWidth( max( maximum_size.width(), (*sub_iter)->rect().left() ) );
                    maximum_size.setHeight( max( maximum_size.height(), (*sub_iter)->rect().bottom() ) );
                }
                
            }
            
        }
        
        // store maximum size and mid position
        if( counts ) _addMidPosition( mid_position/counts + QPoint( _margin(), _margin() ) );
        _addMaximumSize( QSize( maximum_size.width() + 2*_margin(), maximum_size.height() + 2*_margin() ) );
        
    }
    
    // reset index
    _setIndex(0);
    return;
    
}


//______________________________________________________
void TaskBarParabolic::_resize( void )
{
    
    Debug::Throw( "TaskBarParabolic::_resize.\n" );
    
    // if no tasks, set empty message and exit
    if( _buttons().empty() ) {
        resize( emptySize() );
        return;
    }
    
    // dimensions
    _setDimensions();
    _updateMinimumSize();
    _updateMaximumSize();
    _updateMinimumRectPosition();
    
    // svg
    if( _useSvg() && SVG::SvgEngine::get().isValid() ) _preloadSvg();
    
    // make sure taskbar is in minimum state
    // resize
    _setAction( MINIMUM );
    
    resize( _maximumSize() );
    
}

//_________________________________________________________________
void TaskBarParabolic::_placeButtons( void )
{
    
    // compute TaskButtons position
    int x(0);
    int y(0);
    
    const TaskBarAttributes::Layout& layout( attributes().layout() );
    if( layout & (TaskBarAttributes::VERTICAL | TaskBarAttributes::HORIZONTAL ) )
    {
        
        for( TaskButton::List::iterator iter = _buttons().begin(); iter != _buttons().end(); iter++ )
        {
            
            // place button at x and y
            (*iter)->rect().moveTo( x, y );
            
            // calculate index and assign
            (*iter)->setIndex( _index( *iter ) );
            (*iter)->rect().setSize( QSize( (*iter)->currentSize().width() + _spacing(), (*iter)->currentSize().height() + _spacing() ) );
            
            // update x and y position
            if( layout & TaskBarAttributes::VERTICAL ) y+= (*iter)->rect().height();
            else if( layout & TaskBarAttributes::HORIZONTAL ) x += (*iter)->rect().width();
            
        }
        
    } else {
        
        unsigned int column_index(0);
        unsigned int line_index(0);
        
        // compute column width/line height
        vector<unsigned int> column_width( _numColumns(), 0 );
        vector<unsigned int> line_height( _numLines(), 0 );
        for( TaskButton::List::iterator iter = _buttons().begin(); iter != _buttons().end(); iter++ )
        {
            
            (*iter)->setIndex( _index( *iter ) );
            line_height[line_index] = max( line_height[line_index], (*iter)->currentSize().height() + _spacing() );
            column_width[column_index] = max( column_width[column_index], (*iter)->currentSize().width() + _spacing() );
            column_index++;
            if( column_index >= _numColumns() ) {
                column_index = 0;
                line_index++;
            }
            
        }
        
        // place buttons
        column_index = 0;
        line_index = 0;
        for( TaskButton::List::iterator iter = _buttons().begin(); iter != _buttons().end(); iter++ )
        {
            (*iter)->setRect( QRect( QPoint( x, y ), QSize( column_width[column_index], line_height[line_index] ) ) );
            x+=column_width[column_index];
            column_index++;
            if( column_index >= _numColumns() )
            {
                x = 0;
                column_index = 0;
                y+= line_height[line_index];
                line_index++;
            }
        }
        
    }
    
}


//______________________________________________________
void TaskBarParabolic::_addOffsets( void )
{
    
    // get bounding rect for current button sizes
    QRect bounding_rect;
    for( TaskButton::List::iterator iter = _buttons().begin(); iter != _buttons().end(); iter++ )
    {
        bounding_rect.setWidth( max( bounding_rect.width(), (*iter)->rect().right() ) );
        bounding_rect.setHeight( max( bounding_rect.height(), (*iter)->rect().bottom() ) );
    }
    
    // compute offsets
    const TaskBarAttributes::Layout &layout( attributes().layout() );
    
    QPoint offset( _margin(), _margin() );
    if( (layout & TaskBarAttributes::HORIZONTAL) )
    {
        
        if( _mousePosition().x() > _midPosition( _index() ).x() ) offset.setX( (width() - _maximumSize( _index() ).width())/2 + _margin() );
        else offset.setX( ( width() + _maximumSize( _index() ).width())/2 - bounding_rect.width() - _margin());
        
    } else if( (layout & TaskBarAttributes::VERTICAL) ) {
        
        if( _mousePosition().y() > _midPosition( _index() ).y() ) offset.setY( (height() - _maximumSize( _index() ).height())/2 + _margin() );
        else offset.setY( ( height() + _maximumSize( _index() ).height())/2 - bounding_rect.height() - _margin() );
        
    } else {
        
        offset = QPoint( (size().width() - bounding_rect.width())/2, (size().height() - bounding_rect.height())/2 );
        
    }
    
    _setMouseOffset( offset );
    
    // add offsets
    for( TaskButton::List::iterator iter = _buttons().begin(); iter != _buttons().end(); iter++ )
    {
        
        QPoint button_offset( offset );
        if( layout & TaskBarAttributes::VERTICAL )
        {
            
            if( layout & TaskBarAttributes::TOP_LEFT ) button_offset.setX( _margin() );
            else if( layout & TaskBarAttributes::BOTTOM_RIGHT ) button_offset.setX( width() - (*iter)->rect().width() - _margin() );
            else if( layout & TaskBarAttributes::CENTER ) button_offset.setX( (width() - (*iter)->rect().width())/2 );
            
        } else if( layout & TaskBarAttributes::HORIZONTAL ) {
            
            if( layout & TaskBarAttributes::TOP_LEFT ) button_offset.setY( _margin() );
            else if( layout & TaskBarAttributes::BOTTOM_RIGHT ) button_offset.setY( height() - (*iter)->rect().height() - _margin() );
            else if( layout & TaskBarAttributes::CENTER ) button_offset.setY( (height() - (*iter)->rect().height())/2 );
            
        }
        
        (*iter)->setRect( (*iter)->rect().translated( button_offset ) );
        
    }
    
}

//___________________________________________________
bool TaskBarParabolic::_drawButtons( QPaintDevice& device, QRect rect )
{
    
    // if no tasks, do nothing
    if( _buttons().empty() ) return false;
    
    // place the buttons
    _placeButtons();
    _addOffsets();
    
    // redraw
    QPainter painter( &device );
    painter.setClipRect( rect );
    
    if( TRANSPARENCY::CompositeEngine::get().isEnabled() )
    {
        painter.setRenderHints(QPainter::SmoothPixmapTransform);
        painter.setCompositionMode(QPainter::CompositionMode_Source );
    }
    
    // draw root pixmap into entire window
    painter.drawPixmap( 0, 0, _rootPixmap() );
    
    // draw background pixmap clipped to visibleRect
    _updateVisibleRect();
    _drawVisibleRect( painter );
    
    // draw buttons, translated by offset if needed
    painter.setCompositionMode(QPainter::CompositionMode_SourceOver );
    painter.translate( _offset() );
    for( TaskButton::List::iterator iter = _buttons().begin(); iter != _buttons().end(); iter++ )
    { (*iter)->paint( painter ); }
    
    painter.end();
    
    // update index
    unsigned int increment = attributes().flag( TaskBarAttributes::ZOOM_ENABLED ) ? 1:_maximumIndex()-1;
    if( _direction() == TaskButton::BIGGER && _index() < _maximumIndex() -1 ) { if( _action() != ENTER ) _setIndex( _index()+increment ); }
    else if( _direction() == TaskButton::SMALLER && _index() > 0 ) _setIndex( _index()-increment );
    else _setDirection( TaskButton::NONE );
    
    bool out( false );
    for( TaskButton::List::iterator iter = _buttons().begin(); iter != _buttons().end(); iter++ )
    {
        
        if( _action() != ENTER )
        {
            (*iter)->updateHighlightDirection();
            out |= (*iter)->highlightDirection() != TaskButton::NOHIGHLIGHT;
        }
        
    }
    
    return out || _direction() != TaskButton::NONE;
    
}

//___________________________________________________
void TaskBarParabolic::_drawVisibleRect( QPainter& painter )
{
    painter.save();
    
    if( _useSvg() && SVG::SvgEngine::get().isValid() )
    {
        
        painter.translate( _visibleRect().topLeft() );
        painter.drawPixmap( QPoint(0,0), SVG::SvgEngine::get().get( _visibleRect().size() ) );
        
    } else {
        
        if( attributes().flag( TaskBarAttributes::ROUND_CORNERS )  ) painter.setClipRegion( RoundedRegion( _visibleRect(), RoundedRegion::ALL ) );
        else painter.setClipRect( _visibleRect() );
        painter.drawPixmap( 0, 0, _backgroundPixmap() );
        
    }
    painter.restore();
}

//___________________________________________________
void TaskBarParabolic::_setEnableMove( const bool& value )
{
    Debug::Throw( "TaskBar::_setEnableMove.\n" );
    if( _moveEnabled() == value ) return;
    _setDirection( TaskButton::SMALLER );
    TaskBarBase::_setEnableMove( value );
}

//_________________________________________________________________
unsigned int TaskBarParabolic::_index( const TaskButton::Pointer& button )
{
    
    if( !_index() ) return 0;
    
    double distance(0);
    const TaskBarAttributes::Layout& layout( attributes().layout() );
    if( layout & TaskBarAttributes::HORIZONTAL )
    {
        
        distance = fabs(double( _mousePosition().x() - _mouseOffset().x() ) - double( button->rect().left() + button->size(_index()).width()/2 ) );
        distance /= _buttons().size()*( _maximumButtonSize().width() + _spacing() );
        
    } else if( layout & TaskBarAttributes::VERTICAL ) {
        
        distance = fabs( double( _mousePosition().y() - _mouseOffset().y() ) - double( button->rect().top() + button->size(_index()).height()/2 ) );
        distance /= _buttons().size()*( _maximumButtonSize().height() + _spacing() );
        
    } else {
        
        distance = 0;
        if( _numColumns() > 1 ) {
            distance += fabs(double( _mousePosition().x() ) - double( button->rect().left() + button->size(_index()).width()/2 ) )/( _maximumButtonSize().width() + _spacing() )/_numColumns();
        }
        
        if( _numLines() > 1 ) {
            distance += fabs(double( _mousePosition().y() ) - double( button->rect().top() + button->size(_index()).height()/2 ) )/( _maximumButtonSize().height() + _spacing() )/_numLines();
        }
        
    }
    
    return _index( distance );
    
}

//_________________________________________________________________
unsigned int TaskBarParabolic::_index( const double& distance )
{
    
    if( !_index() ) return 0;
    double out( _index()*(1-Util::sqr(attributes().scale()*distance) ) );
    if( out < 0 ) return 0;
    if( out > _maximumIndex()-1 ) return _maximumIndex()-1;
    return (unsigned int) out;
    
}

//_________________________________________________________________
void TaskBarParabolic::_preloadSvg( void ) const
{
    
    Debug::Throw( "TaskBarParabolic::_preloadSvg.\n" );
    assert( _useSvg() && !maximumSizes_.empty() );
    
    // loop over all possible sizes to preload cache
    QSize maximum_size( maximumSizes_.back() );
    QSize minimum_size( minimumRect_.size() );
    
    const TaskBarAttributes::Layout& layout( attributes().layout() );
    SVG::SvgId::List ids;
    if( layout & ( TaskBarAttributes::HORIZONTAL ) )
    {
        
        Debug::Throw() << "TaskBarParabolic::_preloadSvg - (" << minimum_size.width() << "," << maximum_size.width() << ")" << endl;
        
        for( int width = minimum_size.width(); width <= maximum_size.width(); width++ )
        { ids.push_back( QSize( width, minimum_size.height() ) ); }
        
    } else if( layout & (TaskBarAttributes::VERTICAL ) ) {
        
        for( int height = minimum_size.height(); height <= maximum_size.height(); height++ )
        { ids.push_back( QSize( minimum_size.width(), height ) ); }
        
    } else ids.push_back( minimum_size );
    
    SVG::SvgEngine::get().preload( ids );
    
    return;
}
